# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import difflib
import logging
import os
import shutil
import typing as t

import click

from idf_ci._compat import UNDEF

logger = logging.getLogger(__name__)


########################
# Validation Functions #
########################
def _semicolon_separated_list(ctx, param, value):  # noqa: ARG001
    if value is None:
        return None

    if not isinstance(value, str):
        raise click.BadParameter('Value must be a string')

    return [v.strip() for v in value.split(';') if v.strip()]


###########
# Options #
###########
_OPTION_PATHS_HELP = """
List of directories to process. Support passing multiple times.

\b
Example:
    --paths component_1 --paths component_2
    -p component_1 -p component_2
"""


def option_paths(func):
    return click.option(
        '--paths',
        '-p',
        multiple=True,
        type=click.Path(dir_okay=True, file_okay=False, exists=True),
        help=_OPTION_PATHS_HELP,
        callback=lambda ctx, param, value: [os.getcwd()] if not value else value,  # noqa: ARG005
    )(func)


def option_target(func):
    return click.option(
        '--target', '-t', default='all', help='Target to be processed. Or "all" to process all targets.'
    )(func)


def option_parallel(func):
    func = click.option('--parallel-count', default=1, help='Number of parallel builds')(func)
    func = click.option('--parallel-index', default=1, help='Index of the parallel build')(func)
    return func


def option_pytest(func):
    func = click.option('-m', '--marker-expr', default=UNDEF, help='Pytest marker expression, "-m" option')(func)
    func = click.option('-k', '--filter-expr', help='Pytest filter expression, "-k" option')(func)
    return func


def option_modified_files(func):
    return click.option(
        '--modified-files',
        help='Semicolon separated list of files that have been modified',
        callback=_semicolon_separated_list,
    )(func)


def option_commit_sha(func):
    return click.option(
        '--commit-sha',
        help=(
            'Commit SHA to download artifacts from. If not provided, will try to get it from: '
            '1) PIPELINE_COMMIT_SHA env var, 2) latest commit from branch.'
        ),
    )(func)


def option_branch(func):
    return click.option(
        '--branch',
        help='Git branch to use. If not provided, will use current git branch.',
    )(func)


#########
# Utils #
#########
def create_config_file(template_filepath: str, dest: t.Optional[str] = None) -> str:
    """Create a configuration file from a template.

    If the destination file already exists, shows a diff between the existing file and
    the template. If they're identical, notifies the user without making changes.

    :param template_filepath: Path to the template file
    :param dest: Path to the destination file or directory

    :returns: Path to the created or existing file
    """
    if dest is None:
        dest = os.getcwd()

    if os.path.isdir(dest):
        filepath = os.path.join(dest, os.path.basename(template_filepath))
    else:
        filepath = dest

    if not os.path.isfile(filepath):
        shutil.copyfile(template_filepath, filepath)
        click.echo(f'Created {filepath}')
        return filepath

    with open(template_filepath) as template_file:
        template_content = template_file.readlines()
    with open(filepath) as existing_file:
        existing_content = existing_file.readlines()

    diff = list(difflib.unified_diff(existing_content, template_content, fromfile='existing', tofile='template'))
    if not diff:
        click.secho(f'{filepath} already exists and is identical to the template.', fg='yellow')
        return filepath

    click.secho(f'{filepath} already exists. Showing diff:', fg='yellow')
    for line in diff:
        if line.startswith('+'):
            click.secho(line, fg='green', nl=False)
        elif line.startswith('-'):
            click.secho(line, fg='red', nl=False)
        elif line.startswith('@@'):
            click.secho(line, fg='cyan', nl=False)
        else:
            click.secho(line, nl=False)

    return filepath



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import re
import typing as t

import yaml

logger = logging.getLogger(__name__)


def _doublequote_string(value: str) -> str:
    """Double-quote a string if it contains special characters.

    according to https://github.com/motdotla/dotenv, this function is not 100%
    compatible with dotenv, but should work for our use case.
    """
    if any(c in value for c in ' \n'):
        return f'"{value}"'
    return value


def pipeline_variables() -> t.Dict[str, str]:
    """Extract pipeline variables from Gitlab MR predefined variables.

    Possibly set the following variables:

    - IDF_CI_IS_DEBUG_PIPELINE

      Set to '1' if the pipeline is a debug pipeline, will fail at the last stage.

    - IDF_CI_SELECT_ALL_PYTEST_CASES

      Selecting all pytest cases to run

    - IDF_CI_SELECT_BY_FILTER_EXPR

      Build and test only the test cases that match the filter expression (pytest -k)

    - PIPELINE_COMMIT_SHA

      Real commit SHA, instead of the merged result commit SHA

    - INCLUDE_NIGHTLY_RUN

      Run all test cases with or without `nightly_run` marker

    - NIGHTLY_RUN

      Run only test cases with `nightly_run` marker, by default, test cases with
      `nightly_run` marker are skipped
    """
    res: t.Dict[str, str] = {}

    # non-MR pipelines
    if os.getenv('CI_MERGE_REQUEST_IID') is None:
        res['IDF_CI_SELECT_ALL_PYTEST_CASES'] = '1'
        logger.info('Setting `IDF_CI_SELECT_ALL_PYTEST_CASES=1` since running in a non-MR pipeline')

        if os.getenv('CI_COMMIT_SHA'):
            res['PIPELINE_COMMIT_SHA'] = os.environ['CI_COMMIT_SHA']
            logger.info('Setting `PIPELINE_COMMIT_SHA` to `CI_COMMIT_SHA` since running in a non-MR pipeline')
        return {k: _doublequote_string(v) for k, v in res.items()}

    if os.getenv('CI_MERGE_REQUEST_SOURCE_BRANCH_SHA'):
        res['PIPELINE_COMMIT_SHA'] = os.environ['CI_MERGE_REQUEST_SOURCE_BRANCH_SHA']
        logger.info('Setting `PIPELINE_COMMIT_SHA` to `CI_MERGE_REQUEST_SOURCE_BRANCH_SHA`')

    if os.getenv('CI_PYTHON_CONSTRAINT_BRANCH'):
        res['IDF_CI_SELECT_ALL_PYTEST_CASES'] = '1'
        logger.info(
            'Setting `IDF_CI_SELECT_ALL_PYTEST_CASES=1` since pipeline is triggered with a python constraint branch'
        )
    else:
        mr_labels = os.getenv('CI_MERGE_REQUEST_LABELS', '').split(',')
        if 'include_nightly_run' in mr_labels:
            res['INCLUDE_NIGHTLY_RUN'] = '1'
            logger.info('Setting `INCLUDE_NIGHTLY_RUN=1` since MR label `include_nightly_run` is set')
            res['NIGHTLY_RUN'] = '1'
            logger.info('Setting `NIGHTLY_RUN=1` since MR label `include_nightly_run` is set')

        if 'nightly' in mr_labels:
            res['NIGHTLY_RUN'] = '1'
            logger.info('Setting `NIGHTLY_RUN=1` since MR label `nightly` is set')

        if 'BUILD_AND_TEST_ALL_APPS' in mr_labels:
            res['IDF_CI_SELECT_ALL_PYTEST_CASES'] = '1'
            logger.info('Setting `IDF_CI_SELECT_ALL_PYTEST_CASES=1` since MR label `BUILD_AND_TEST_ALL_APPS` is set')
        else:
            description = os.getenv('CI_MERGE_REQUEST_DESCRIPTION', '')
            if description:
                pattern = r'^## Dynamic Pipeline Configuration(?:[^`]*?)```(?:\w+)(.*?)```'
                result = re.search(pattern, description, re.DOTALL | re.MULTILINE)
                if result:
                    data = yaml.safe_load(result.group(1))
                    if 'Test Case Filters' in data:
                        res['IDF_CI_SELECT_BY_FILTER_EXPR'] = ' or '.join(data.get('Test Case Filters'))
                        logger.info(
                            f'Setting `IDF_CI_SELECT_BY_FILTER_EXPR={res["IDF_CI_SELECT_BY_FILTER_EXPR"]}` '
                            f'based on MR description "Test Case Filters"'
                        )
                        res['IDF_CI_IS_DEBUG_PIPELINE'] = '1'
                        logger.info('Setting `IDF_CI_IS_DEBUG_PIPELINE=1` based on MR description "Test Case Filters"')

    return {k: _doublequote_string(v) for k, v in res.items()}



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os

import pytest

from idf_ci.idf_gitlab.scripts import pipeline_variables


class TestPipelineVariables:
    @pytest.fixture(autouse=True)
    def setup_clean_env(self, monkeypatch):
        for env_var in [var for var in os.environ if var.startswith(('CI_', 'IDF_CI_'))]:
            monkeypatch.delenv(env_var, raising=False)

    def test_non_mr_pipeline(self, monkeypatch):
        monkeypatch.setenv('CI_COMMIT_SHA', '12345abcde')

        assert pipeline_variables() == {
            'IDF_CI_SELECT_ALL_PYTEST_CASES': '1',
            'PIPELINE_COMMIT_SHA': '12345abcde',
        }

    def test_mr_python_constraint(self, monkeypatch):
        monkeypatch.setenv('CI_MERGE_REQUEST_IID', '123')
        monkeypatch.setenv('CI_COMMIT_SHA', 'bcdefa54321')
        monkeypatch.setenv('CI_MERGE_REQUEST_SOURCE_BRANCH_SHA', 'abcdef12345')
        monkeypatch.setenv('CI_PYTHON_CONSTRAINT_BRANCH', 'some-branch')

        assert pipeline_variables() == {
            'IDF_CI_SELECT_ALL_PYTEST_CASES': '1',
            'PIPELINE_COMMIT_SHA': 'abcdef12345',
        }

    def test_mr_build_all_apps(self, monkeypatch):
        monkeypatch.setenv('CI_MERGE_REQUEST_IID', '123')
        monkeypatch.setenv('CI_COMMIT_SHA', 'bcdefa54321')
        monkeypatch.setenv('CI_MERGE_REQUEST_SOURCE_BRANCH_SHA', 'abcdef12345')
        monkeypatch.setenv('CI_MERGE_REQUEST_LABELS', 'BUILD_AND_TEST_ALL_APPS,some-other-label')

        assert pipeline_variables() == {
            'IDF_CI_SELECT_ALL_PYTEST_CASES': '1',
            'PIPELINE_COMMIT_SHA': 'abcdef12345',
        }

    def test_mr_test_filters(self, monkeypatch):
        monkeypatch.setenv('CI_MERGE_REQUEST_IID', '123')
        monkeypatch.setenv('CI_COMMIT_SHA', 'bcdefa54321')
        monkeypatch.setenv('CI_MERGE_REQUEST_SOURCE_BRANCH_SHA', 'abcdef12345')
        monkeypatch.setenv(
            'CI_MERGE_REQUEST_DESCRIPTION',
            '## Dynamic Pipeline Configuration\n\n'
            '```yaml\n'
            'Test Case Filters:\n'
            '  - filter1\n'
            '  - filter2\n'
            '```\n\n'
            'Some other text',
        )

        assert pipeline_variables() == {
            'IDF_CI_SELECT_BY_FILTER_EXPR': '"filter1 or filter2"',
            'IDF_CI_IS_DEBUG_PIPELINE': '1',
            'PIPELINE_COMMIT_SHA': 'abcdef12345',
        }

    def test_mr_no_special_conditions(self, monkeypatch):
        monkeypatch.setenv('CI_MERGE_REQUEST_IID', '123')
        monkeypatch.setenv('CI_COMMIT_SHA', 'bcdefa54321')
        monkeypatch.setenv('CI_MERGE_REQUEST_SOURCE_BRANCH_SHA', 'abcdef12345')

        assert pipeline_variables() == {'PIPELINE_COMMIT_SHA': 'abcdef12345'}

    def test_no_env_vars(self):
        assert pipeline_variables() == {'IDF_CI_SELECT_ALL_PYTEST_CASES': '1'}



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

__all__ = [
    'ArtifactManager',
    'ArtifactParams',
    'build_child_pipeline',
    'pipeline_variables',
    'test_child_pipeline',
]


from .api import ArtifactManager, ArtifactParams
from .pipeline import build_child_pipeline, test_child_pipeline
from .scripts import pipeline_variables



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0


__all__ = [
    'IDF_CI_PLUGIN_KEY',
    'IDF_CI_PYTEST_CASE_KEY',
    'IDF_CI_PYTEST_DEBUG_INFO_KEY',
    'GroupedPytestCases',
    'IdfPytestPlugin',
    'PytestApp',
    'PytestCase',
    'get_pytest_cases',
]

from idf_ci.idf_pytest.models import GroupedPytestCases, PytestApp, PytestCase
from idf_ci.idf_pytest.plugin import (
    IDF_CI_PLUGIN_KEY,
    IDF_CI_PYTEST_CASE_KEY,
    IDF_CI_PYTEST_DEBUG_INFO_KEY,
    IdfPytestPlugin,
)
from idf_ci.idf_pytest.scripts import get_pytest_cases



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

__all__ = ['click_cli']


import logging
import os
import typing as t
from ast import literal_eval

import click

from idf_ci.cli._options import create_config_file
from idf_ci.cli.build_group import build
from idf_ci.cli.gitlab_group import gitlab
from idf_ci.cli.test_group import test
from idf_ci.settings import CiSettings
from idf_ci.utils import setup_logging

logger = logging.getLogger(__name__)


@click.group(context_settings={'show_default': True, 'help_option_names': ['-h', '--help']})
@click.option(
    '-c',
    '--config-file',
    type=click.Path(dir_okay=False, file_okay=True, exists=True),
    help='Path to the idf-ci config file',
)
@click.option(
    '--config',
    type=str,
    multiple=True,
    help=(
        'Override settings via dot-path assignments (repeatable). '
        'Format: path.to.key = value. '
        'Values use Python literal syntax (ast.literal_eval): 10, True, "str", {..}, [..]. '
        'Precedence: CLI > init > config file > defaults'
    ),
)
@click.option('--debug', is_flag=True, default=False, help='Enable debug logging')
def click_cli(config_file, config, debug):
    """ESP-IDF CI CLI Tool."""
    if debug:
        setup_logging(logging.DEBUG)
    else:
        setup_logging()

    if config_file:
        logger.debug(f'Using config file: {config_file}')
        CiSettings.CONFIG_FILE_PATH = config_file

    if config:

        def _set_nested(target: t.Dict, path: str, value: t.Any):
            parts = [p.strip() for p in path.split('.') if p.strip()]
            if not parts:
                raise click.BadParameter('Empty key in --config assignment')

            cursor = target
            for key in parts[:-1]:  # check all but last
                if key not in cursor or not isinstance(cursor.get(key), dict):
                    cursor[key] = {}
                cursor = cursor[key]
            cursor[parts[-1]] = value

        overrides_dict = {}
        for item in config:  # item like: "a.b = 1"
            left, sep, right = item.partition('=')
            if not sep:
                raise click.BadParameter(f'Invalid --config entry `{item}`. Expected format: `path.to.key = value`')
            try:
                value = literal_eval(right.strip())
            except Exception as e:
                raise click.BadParameter(f'Failed to parse value in --config `{item}`: {e}')

            _set_nested(overrides_dict, left.strip(), value)

        # Merge with existing overrides if any
        def _deep_merge(dst: t.Dict, src: t.Dict):
            for k, v in src.items():
                if isinstance(v, dict) and isinstance(dst.get(k), dict):
                    _deep_merge(dst[k], v)
                else:
                    dst[k] = v
            return dst

        _deep_merge(CiSettings.CLI_OVERRIDES, overrides_dict)


@click_cli.command()
@click.option('--path', help='Path to create the config file')
def init(path: str):
    """Create .idf_ci.toml with default values."""
    create_config_file(os.path.join(os.path.dirname(__file__), '..', 'templates', '.idf_ci.toml'), path)


@click_cli.command()
def completions():
    """Instructions to enable shell completions for idf-ci."""
    click.echo("""
    To enable autocomplete run the following command:

    Bash:
      1. Run this command once

        _IDF_CI_COMPLETE=bash_source idf-ci > ~/.idf-ci-complete.bash

      2. Add the following line to your .bashrc
        . ~/.idf-ci-complete.bash

    Zsh:
      1. Run this command once

        _IDF_CI_COMPLETE=zsh_source idf-ci > ~/.idf-ci-complete.zsh

      2. Add the following line to your .zshrc

        . ~/.idf-ci-complete.zsh

    Fish:
      1. Run this command once

        _IDF_CI_COMPLETE=fish_source idf-ci > ~/.config/fish/completions/idf-ci.fish

    After modifying the shell config, you need to start a new shell in order for the changes to be loaded.
    """)


click_cli.add_command(build)
click_cli.add_command(test)
click_cli.add_command(gitlab)



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
from pathlib import Path

import pytest

from idf_ci.cli import click_cli
from idf_ci.settings import CiSettings


@pytest.mark.parametrize(
    'command, default_file, specific_file',
    [
        (['build', 'init'], '.idf_build_apps.toml', 'custom_build.toml'),
        (['init'], '.idf_ci.toml', 'custom_ci.toml'),
        (['test', 'init'], 'pytest.ini', 'custom_test.ini'),
    ],
)
def test_init_commands(runner, tmp_dir, command, default_file, specific_file):
    # Test init command with default path
    with runner.isolated_filesystem():
        result = runner.invoke(click_cli, [*command, '--path', tmp_dir])
        assert result.exit_code == 0
        assert f'Created {os.path.join(tmp_dir, default_file)}' in result.output
        assert os.path.exists(os.path.join(tmp_dir, default_file))

    # Test init command with specific file path
    specific_path = os.path.join(tmp_dir, specific_file)
    result = runner.invoke(click_cli, [*command, '--path', specific_path])
    assert result.exit_code == 0
    assert f'Created {specific_path}' in result.output
    assert os.path.exists(specific_path)


def test_completions(runner):
    result = runner.invoke(click_cli, ['completions'])
    assert result.exit_code == 0
    assert 'To enable autocomplete run the following command:' in result.output
    assert 'Bash:' in result.output
    assert 'Zsh:' in result.output
    assert 'Fish:' in result.output


def test_init_but_already_exists(runner, tmp_dir):
    build_profile_path = os.path.join(tmp_dir, '.idf_build_apps.toml')
    ci_profile_path = os.path.join(tmp_dir, '.idf_ci.toml')

    # Create files first
    Path(build_profile_path).touch()
    Path(ci_profile_path).touch()

    # Try to init again
    result = runner.invoke(click_cli, ['build', 'init', '--path', tmp_dir])
    assert result.exit_code == 0

    result = runner.invoke(click_cli, ['init', '--path', tmp_dir])
    assert result.exit_code == 0

    result = runner.invoke(click_cli, ['test', 'init', '--path', tmp_dir])
    assert result.exit_code == 0


class TestConfig:
    @pytest.fixture(autouse=True)
    def _setup_method(self):
        CiSettings.CLI_OVERRIDES = {}

    @pytest.fixture(autouse=True)
    def _teardown_method(self):
        CiSettings.CLI_OVERRIDES = {}

    def test_config_nested_assignment_with_spaces(self, runner):
        result = runner.invoke(
            click_cli,
            ['--config', 'gitlab.build_pipeline.runs_per_job = 10', 'completions'],
        )
        assert result.exit_code == 0

        settings = CiSettings()
        assert settings.gitlab.build_pipeline.runs_per_job == 10

    def test_config_multiple_overrides_merged(self, runner):
        result = runner.invoke(
            click_cli,
            [
                '--config',
                'gitlab.build_pipeline.runs_per_job=10',
                '--config',
                'gitlab.build_pipeline.workflow_name="A"',
                'completions',
            ],
        )
        assert result.exit_code == 0

        settings = CiSettings()
        assert settings.gitlab.build_pipeline.runs_per_job == 10
        assert settings.gitlab.build_pipeline.workflow_name == 'A'

    def test_config_invalid_value_errors(self, runner):
        result = runner.invoke(
            click_cli,
            ['--config', 'gitlab.build_pipeline.runs_per_job = not_a_literal', 'completions'],
        )
        assert result.exit_code != 0
        assert 'Failed to parse value in --config `gitlab.build_pipeline.runs_per_job = not_a_literal`' in result.output

    def test_config_invalid_format_errors(self, runner):
        result = runner.invoke(click_cli, ['--config', 'abc', 'completions'])
        assert result.exit_code != 0
        assert 'Invalid --config entry `abc`' in result.output

    def test_config_empty_key_errors(self, runner):
        result = runner.invoke(click_cli, ['--config', ' = 1', 'completions'])
        assert result.exit_code != 0
        assert 'Empty key in --config assignment' in result.output
